<?php declare(strict_types=1);

namespace Addons\Reports\Feedback\Reports\Cards\Responses;

use Addons\Reports\Feedback\Controllers\ModelConcerns;
use App\Modules\Core\Models\CustomField;
use App\Modules\Core\Models\FeedbackFormField;
use App\Modules\Core\Models\FeedbackFormFieldValue;
use App\Modules\Core\Models\FeedbackLog;
use App\Modules\Report\Addon\Dashboards\Cards\Table\Column;
use App\Modules\Report\Addon\Dashboards\Cards\TableCard;
use App\Modules\Report\Addon\Dashboards\Filtering\QueryBuilder;
use App\Modules\Selfservice\Models\Article;
use App\Modules\Ticket\Models\Ticket;
use Form;

use function e;
use function get_class;
use function route;
use function trans;

/**
 * @template T of Ticket|Article
 * @template TFeedbackLog of FeedbackLog
 * @template TFeedbackFormFieldValue of FeedbackFormFieldValue
 * @extends TableCard<T>
 */
abstract class FieldTable extends TableCard
{
    /** @phpstan-use ModelConcerns<T, TFeedbackLog, TFeedbackFormFieldValue> */
    use ModelConcerns;

    protected FeedbackFormField $field;

    /** @var string[] */
    protected array $with = [];

    public function __construct(FeedbackFormField $field)
    {
        parent::__construct();

        $this->field = $field;
    }

    public function name(): string
    {
        return trans('report.feedback_field', ['name' => $this->field->name]);
    }

    public function query(): QueryBuilder
    {
        /** @var class-string<T> $model */
        $model = get_class($this->model());
        $filters = $this->dashboard()->filters($this->getQualifiedName($this->logModel()->getTable(), 'updated_at'));

        return QueryBuilder::for($model)
            ->applyFilters($filters)
            ->join(
                $this->logModel()->getTable(),
                $this->logRelation()->getQualifiedParentKeyName(),
                '=',
                $this->logRelation()->getQualifiedForeignKeyName()
            )
            ->leftJoin(
                $this->valueModel()->getTable(),
                $this->getQualifiedName($this->logModel()->getTable(), 'id'),
                '=',
                $this->valueRelation()->getQualifiedForeignKeyName()
            )
            ->where(
                $this->getQualifiedName($this->valueModel()->getTable(), '.field_id'),
                $this->field->id
            )
            ->with($this->with)
            ->with([
                'feedback.values' => function ($query) {
                    $query->where('field_id', $this->field->id);
                }
            ])
            ->select($this->getQualifiedName($this->model()->getTable(), '*'));
    }

    protected function dateColumn(): Column
    {
        return (new Column(trans('general.date')))
            ->setId('updated_at')
            ->setValue(fn($model) => formatDate($model->feedback->updated_at));
    }

    protected function ratingColumn(): Column
    {
        return (new Column(trans('ticket.rating')))
            ->setValue(fn($model) => $model->feedback?->rating?->toString());
    }

    protected function valueColumn(): Column
    {
        return (new Column(trans('general.value')))
            ->setValue(function ($model) {
                $value = $model->feedback->values->firstOrFail()->value;

                if ($this->field->type === CustomField::DATE) {
                    $value = formatDate($value, true);
                }

                return $value;
            });
    }

    protected function actionsColumn(string $name): Column
    {
        return (new Column(''))
            ->setId('actions')
            ->setCsvValue(fn() => '')
            ->setValue(function ($model) use ($name) {
                $button = null;
                if ($model->feedback->fields_answered) {
                    $url = e(route($name, [$model->feedback->id]));
                    $button = Form::button(
                        trans('ticket.view_feedback'),
                        [
                            'class' => 'view-feedback sp-button-sm sp-float-right sp-whitespace-nowrap',
                            'onclick' => "window.open('".$url."','width=800,height=600,scrollbars=yes');"
                        ]
                    )->toHtml();
                }

                return $button;
            }, false);
    }
}
