<?php declare(strict_types=1);

namespace Addons\Reports\Feedback\Reports\Cards\Responses;

use App\Http\Requests\Request;
use App\Modules\Core\Models\CustomField;
use App\Modules\Core\Models\FeedbackFormFieldOption;
use App\Modules\Core\Models\FeedbackFormFieldValue;
use App\Modules\Core\Models\FeedbackLog;
use App\Modules\Core\Models\TranslationCollection;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\ChartType;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\Datasets\PrimitiveDataset;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\DataStructure;
use App\Modules\Selfservice\Models\Article;
use App\Modules\Ticket\Models\Ticket;
use Illuminate\Database\Eloquent\Builder;
use UnhandledMatchError;

use function array_column;
use function escape_like;
use function json_decode;
use function sprintf;

/**
 * @template TModel of Ticket|Article
 * @template TFeedbackLog of FeedbackLog
 * @template TFeedbackFormFieldValue of FeedbackFormFieldValue
 * @template TFeedbackFormFieldOption of FeedbackFormFieldOption
 * @extends FieldChartCard<TModel, TFeedbackLog, TFeedbackFormFieldValue>
 */
abstract class OptionsFieldChart extends FieldChartCard
{
    public function data(): DataStructure
    {
        $results = $this->getChartJsDataset();

        return (new DataStructure(ChartType::BAR, new PrimitiveDataset(array_column($results, 'value'))))
            ->setLabels(array_column($results, 'label'))
            ->setOptions([
                'plugins' => [
                    'legend' => ['display' => false]
                ],
                'scales' => [
                    'y' => [
                        'ticks' => ['precision' => 0]
                    ],
                ],
            ]);
    }

    /**
     * @inheritDoc
     */
    public function underlyingDataQuery(Request $request): Builder
    {
        $query = $this->baseQuery();
        if ($this->isDrillIntoChartRequest($request)) {
            $matched = false;
            $label = $request->string('label')->toString();

            foreach ($this->getFieldOptions() as $option) {
                if ($label !== $option->value) {
                    continue;
                }

                $value = $option->id;
                $operator = '=';

                if ($this->field->type === CustomField::CHECKLIST || $this->field->type === CustomField::MULTIPLE) {
                    $value = '%"' . escape_like((string) $value) . '"%';
                    $operator = 'LIKE';
                }

                $query->where('value', $operator, $value);

                $matched = true;
                break;
            }

            if (! $matched) {
                throw new UnhandledMatchError(sprintf('Unsupported label \'%s\'.', $label));
            }
        }

        return $query->with('feedback')
            ->selectRaw($this->getQualifiedName($this->model()->getTable(), '*'));
    }

    /**
     * @return TranslationCollection<int, TFeedbackFormFieldOption>
     */
    abstract protected function getFieldOptions(): TranslationCollection;

    /**
     * @return mixed[]
     */
    private function getChartJsDataset(): array
    {
        $options = $this->getFieldOptions();

        $results = [];
        foreach ($options as $option) {
            $results[$option->id] = [
                'label' => $option->value,
                'value' => 0
            ];
        }

        $values = $this->baseQuery()
            ->select('value')
            ->getQuery()
            ->get();

        foreach ($values as $value) {
            $value = $value->value;

            if ($this->field->type === CustomField::CHECKLIST || $this->field->type === CustomField::MULTIPLE) {
                $value = json_decode($value);
            }

            foreach ((array) $value as $selected) {
                $results[$selected]['value']++;
            }
        }

        return $results;
    }
}
