<?php declare(strict_types=1);

namespace Addons\Reports\TicketPerformance\Reports\Cards\FirstResponseTime;

use App\Http\Requests\Request;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\ChartType;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\Datasets\CustomObjectDataset;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\DataStructure;
use App\Modules\Report\Addon\Dashboards\Filtering\QueryBuilder;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketPriorityColumn;
use App\Modules\Report\Metrics\Metric;
use App\Modules\Ticket\Models\Ticket;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Support\Collection;

use function array_merge;
use function trans;
use function trans_choice;

class ByPriorityChart extends BaseChart
{
    public function name(): string
    {
        return trans('report.first_response_time_by_priority');
    }

    public function data(): DataStructure
    {
        $data = $this->getChartJsDataset();
        $dataset = new CustomObjectDataset($data->map(fn($v) => (array) $v)->all(), 'average', 'name');

        return (new DataStructure(ChartType::BAR, $dataset))
            ->setLabels($data->pluck('name')->toArray())
            ->setAxisNames(trans_choice('general.hour', 2), trans_choice('ticket.priority', 1))
            ->setOptions([
                'indexAxis' => 'y',
                'plugins' => [
                    'legend' => ['display' => false]
                ],
            ]);
    }

    public function getUnderlyingDataValueKey(): string
    {
        return 'priority_id';
    }

    /**
     * @return Builder<Ticket>
     */
    public function underlyingDataQuery(Request $request): Builder
    {
        $query = $this->baseQuery()
            ->with(['department', 'priority', 'user'])
            ->select(['ticket.*'])
            ->selectRaw('CAST(report_metric_value.value as unsigned) as `first_response_time`');

        if ($this->isDrillIntoChartRequest($request)) {
            $query->where('ticket.priority_id', $request->integer('value'));
        }

        return $query;
    }

    public function underlyingDataTableColumns(): array
    {
        return array_merge(parent::underlyingDataTableColumns(), [new TicketPriorityColumn]);
    }

    /**
     * @return Collection<int,Ticket>
     */
    private function getChartJsDataset(): Collection
    {
        return $this->baseQuery()
            ->groupBy('ticket.priority_id')
            ->selectRaw('ticket.priority_id')
            ->selectRaw('ticket_priority.name')
            ->selectRaw('ROUND(AVG(report_metric_value.value / 3600), 2) as `average`')
            ->getQuery()
            ->get();
    }

    /**
     * @return QueryBuilder<Ticket>
     */
    private function baseQuery(): QueryBuilder
    {
        return QueryBuilder::for(Ticket::class)
            ->applyFilters($this->dashboard()->filters('ticket.created_at'))
            ->join('ticket_priority', 'ticket.priority_id', '=', 'ticket_priority.id')
            ->join('report_metric_value', 'report_metric_value.model_id', '=', 'ticket.id')
            ->join('report_metric', 'report_metric_value.metric_id', '=', 'report_metric.id')
            ->where('report_metric.name', '=', Metric::FIRST_RESPONSE_TIME->value)
            ->orderBy('ticket_priority.order');
    }

    protected function isDrillIntoChartRequest(Request $request): bool
    {
        return $request->has('label') && $request->has('value');
    }
}
