<?php declare(strict_types=1);

namespace Addons\Reports\TicketPerformance\Reports\Cards\FirstResponseTime;

use App\Http\Requests\Request;
use App\Modules\Report\Addon\Concerns\InteractsWithTimeframe;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\ChartType;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\Datasets\ObjectDataset;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\DataStructure;
use App\Modules\Report\Addon\Dashboards\Filtering\QueryBuilder;
use App\Modules\Report\Metrics\Metric;
use App\Modules\Ticket\Models\Ticket;
use Carbon\CarbonInterface;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Support\Collection;

use function get_object_vars;
use function sprintf;
use function trans;
use function trans_choice;

class TrendChart extends BaseChart
{
    use InteractsWithTimeframe;

    protected int $width = 8;

    public function name(): string
    {
        return trans('report.first_response_time_average_trend');
    }

    public function data(): DataStructure
    {
        $unit = $this->calculateUnits();
        $dataset = new ObjectDataset($this->getChartJsDataset()->values()->toArray());

        return (new DataStructure(ChartType::BAR, $dataset))
            ->setAxisNames($this->unitToTrans($unit), trans_choice('general.hour', 2))
            ->setOptions([
                'plugins' => [
                    'legend' => ['display' => false]
                ],
                'scales' => [
                    'x' => [
                        'type'    => 'time',
                        'time'    => [
                            'unit' => $unit,
                            'round' => $unit,
                            'tooltipFormat' => $this->unitToTooltipFormat($unit),
                        ],
                    ],
                    'y' => ['min'   => 0,],
                ]
            ]);
    }

    /**
     * @return Builder<Ticket>
     */
    public function underlyingDataQuery(Request $request): Builder
    {
        $query = $this->baseQuery()
            ->with(['department', 'user'])
            ->select(['ticket.*'])
            ->selectRaw('CAST(report_metric_value.value as unsigned) as `first_response_time`');

        if ($this->isDrillIntoChartRequest($request)) {
            $time = $request->integer('value.x') / 1000;

            $query->whereRaw(sprintf('%s = ?', $this->getTimeSeriesExpr($this->timeframeColumn())), $time);
        }

        return $query;
    }

    /**
     * @return Collection<int,mixed>
     */
    private function getChartJsDataset(): Collection
    {
        $collection = $this->baseQuery()
            ->selectRaw(sprintf('%s * 1000 as x', $this->getTimeSeriesExpr($this->timeframeColumn())))
            ->selectRaw('AVG(report_metric_value.value) / 3600.0 as y')
            ->groupBy('x')
            ->getQuery()
            ->get()
            ->keyBy('x');

        return $this->createDateRange($this->calculateUnits())
            ->map(function (CarbonInterface $d) use ($collection) {
                if (isset($collection[$d->unix() * 1000])) {
                    return get_object_vars($collection[$d->unix() * 1000]);
                }

                return ['x' => $d->unix() * 1000, 'y' => 0];
            });
    }

    /**
     * @return QueryBuilder<Ticket>
     */
    private function baseQuery(): QueryBuilder
    {
        return QueryBuilder::for(Ticket::class)
            ->applyFilters($this->dashboard()->filters($this->timeframeColumn()))
            ->join('report_metric_value', 'report_metric_value.model_id', '=', 'ticket.id')
            ->join('report_metric', 'report_metric_value.metric_id', '=', 'report_metric.id')
            ->where('report_metric.name', '=', Metric::FIRST_RESPONSE_TIME->value);
    }

    private function timeframeColumn(): string
    {
        return 'ticket.created_at';
    }

    protected function isDrillIntoChartRequest(Request $request): bool
    {
        return $request->has('label') && $request->has('value.x') && $request->has('value.y');
    }
}
