<?php declare(strict_types=1);

namespace Addons\Reports\TicketPerformance\Reports\Cards\SlaPerformance;

use App\Http\Requests\Request;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\ChartType;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\Datasets\CustomObjectDataset;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\DataStructure;
use App\Modules\Report\Addon\Dashboards\Cards\ChartCard;
use App\Modules\Report\Addon\Dashboards\Cards\UnderlyingData;
use App\Modules\Report\Addon\Dashboards\Filtering\QueryBuilder;
use App\Modules\Ticket\Models\Ticket;
use App\Modules\Ticket\Models\TicketDueTimeTarget;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Support\Collection;

use function sprintf;
use function trans;
use function trans_choice;

/**
 * @implements UnderlyingData<Ticket>
 */
abstract class ByPriorityChart extends ChartCard implements UnderlyingData
{
    use BaseQuery, HasColumns;

    protected int $height = 3;

    public function name(): string
    {
        return trans('report.first_response_time_by_priority');
    }

    public function data(): DataStructure
    {
        $data = $this->getChartJsDataset();
        $dataset = new CustomObjectDataset($data->map(fn($v) => (array) $v)->all(), 'percentage', 'name');

        return (new DataStructure(ChartType::BAR, $dataset))
            ->setLabels($data->pluck('name')->toArray())
            ->setAxisNames('%', trans_choice('ticket.priority', 1))
            ->setOptions([
                'indexAxis' => 'y',
                'plugins' => [
                    'legend' => ['display' => false]
                ],
                'scales' => [
                    'x' => [
                        'max'   => 100,
                        'min'   => 0,
                        'ticks' => ['precision' => 0],
                    ],
                ],
            ]);
    }

    public function getUnderlyingDataValueKey(): string
    {
        return 'priority_id';
    }

    /**
     * @return Builder<Ticket>
     */
    public function underlyingDataQuery(Request $request): Builder
    {
        $table = TicketDueTimeTarget::getTableName();

        $query = $this->baseQuery()
            ->with('priority')
            ->orderBy(sprintf('%s.created_at', $table))
            ->select('ticket.*', sprintf('%s.target_met', $table), sprintf('%s.breach_time', $table));

        if ($this->isDrillIntoChartRequest($request)) {
            $query->where('ticket.priority_id', $request->integer('value'));
        }

        return $query;
    }

    /**
     * @return Collection<int,mixed>
     */
    private function getChartJsDataset(): Collection
    {
        return $this->priorityBaseQuery()
            ->groupBy('ticket.priority_id')
            ->selectRaw('ticket.priority_id')
            ->selectRaw('ticket_priority.name')
            ->selectRaw(sprintf('count(CASE when %s.target_met = 1 THEN 1 END) / count(*) * 100 as percentage', TicketDueTimeTarget::getTableName()))
            ->getQuery()
            ->get();
    }

    /**
     * @return QueryBuilder<Ticket>
     */
    protected function priorityBaseQuery(): QueryBuilder
    {
        return $this->baseQuery()
            ->join('ticket_priority', 'ticket.priority_id', '=', 'ticket_priority.id')
            ->orderBy('ticket_priority.order');
    }

    protected function isDrillIntoChartRequest(Request $request): bool
    {
        return $request->has('label') && $request->has('value');
    }
}
