<?php declare(strict_types=1);

namespace Addons\Reports\TicketPerformance\Reports\Cards\TimeToResolution;

use App\Http\Requests\Request;
use App\Modules\Report\Addon\Concerns\UnixTimestamp;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\ChartType;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\Datasets\CustomObjectDataset;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\DataStructure;
use App\Modules\Report\Addon\Dashboards\Cards\ChartCard;
use App\Modules\Report\Addon\Dashboards\Cards\Table\Column;
use App\Modules\Report\Addon\Dashboards\Cards\UnderlyingData;
use App\Modules\Report\Addon\Dashboards\Filtering\QueryBuilder;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketCreatedAtColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketDepartmentColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketNumberColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketPriorityColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketSubjectColumn;
use App\Modules\Report\Metrics\Metric;
use App\Modules\Ticket\Models\Ticket;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\Config;

use function trans;
use function trans_choice;

/**
 * @implements UnderlyingData<Ticket>
 */
class ByPriorityChart extends ChartCard implements UnderlyingData
{
    use UnixTimestamp;

    protected int $width = 6;

    public function name(): string
    {
        return trans('report.time_resolution_by_priority');
    }

    public function data(): DataStructure
    {
        $data = $this->getChartJsDataset();
        $dataset = new CustomObjectDataset($data->map(fn($v) => (array) $v)->all(), 'average', 'name');

        return (new DataStructure(ChartType::BAR, $dataset))
            ->setLabels($data->pluck('name')->toArray())
            ->setAxisNames(trans_choice('general.hour', 2), trans_choice('ticket.priority', 1))
            ->setOptions([
                'indexAxis' => 'y',
                'plugins' => [
                    'legend' => ['display' => false]
                ],
            ]);
    }

    public function getUnderlyingDataValueKey(): string
    {
        return 'priority_id';
    }

    /**
     * @return Builder<Ticket>
     */
    public function underlyingDataQuery(Request $request): Builder
    {
        $query = $this->baseQuery()
            ->with('department', 'priority')
            ->select(['ticket.*'])
            ->selectRaw('CAST(report_metric_value.value as unsigned) as `time_to_resolution`');

        if ($this->isDrillIntoChartRequest($request)) {
            $query->where('ticket.priority_id', $request->integer('value'));
        }

        return $query;
    }

    public function underlyingDataTableColumns(): array
    {
        $timeColumn = (new Column(trans('report.time_to_resolution')))
            ->setId('time_to_resolution')
            ->setDefinition(['width' => '160px'])
            ->setValue(fn($ticket) => $this->formatTime((int) $ticket->time_to_resolution));

        return [
            $timeColumn,
            new TicketCreatedAtColumn,
            new TicketNumberColumn,
            new TicketSubjectColumn,
            new TicketDepartmentColumn,
            new TicketPriorityColumn,
        ];
    }

    /**
     * @return Collection<int,Ticket>
     */
    private function getChartJsDataset(): Collection
    {
        return $this->baseQuery()
            ->groupBy('ticket.priority_id')
            ->selectRaw('ticket.priority_id')
            ->selectRaw('ticket_priority.name')
            ->selectRaw('ROUND(AVG(report_metric_value.value / 3600), 2) as `average`')
            ->getQuery()
            ->get();
    }

    /**
     * @return QueryBuilder<Ticket>
     */
    private function baseQuery(): QueryBuilder
    {
        return QueryBuilder::for(Ticket::class)
            ->applyFilters($this->dashboard()->filters('ticket.resolved_time'))
            ->join('report_metric_value', 'report_metric_value.model_id', '=', 'ticket.id')
            ->join('report_metric', 'report_metric_value.metric_id', '=', 'report_metric.id')
            ->join('ticket_priority', 'ticket.priority_id', '=', 'ticket_priority.id')
            ->where('ticket.status_id', Config::get('settings.default_resolved_status'))
            ->where('report_metric.name', '=', Metric::TIME_TO_RESOLUTION->value)
            ->orderBy('ticket_priority.order');
    }

    protected function isDrillIntoChartRequest(Request $request): bool
    {
        return $request->has('label') && $request->has('value');
    }
}
