<?php declare(strict_types=1);

namespace Addons\Reports\TicketPerformance\Reports\Cards\TimeToResolution;

use App\Http\Requests\Request;
use App\Modules\Report\Addon\Concerns\UnixTimestamp;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\ChartType;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\Datasets\PrimitiveDataset;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\DataStructure;
use App\Modules\Report\Addon\Dashboards\Cards\ChartCard;
use App\Modules\Report\Addon\Dashboards\Cards\Table\Column;
use App\Modules\Report\Addon\Dashboards\Cards\UnderlyingData;
use App\Modules\Report\Addon\Dashboards\Filtering\QueryBuilder;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketCreatedAtColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketDepartmentColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketNumberColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketSubjectColumn;
use App\Modules\Report\Metrics\Metric;
use App\Modules\Ticket\Models\Ticket;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Support\Facades\Config;
use UnhandledMatchError;

use function array_values;
use function sprintf;
use function trans;

/**
 * @implements UnderlyingData<Ticket>
 */
class ByTimeGraph extends ChartCard implements UnderlyingData
{
    use UnixTimestamp;

    protected int $width = 6;

    public function name(): string
    {
        return trans('report.time_to_resolution');
    }

    public function data(): DataStructure
    {
        $dataset = new PrimitiveDataset($this->getChartJsDataset());

        return (new DataStructure(ChartType::BAR, $dataset))
            ->setLabels([
                trans('report.0-6_hours'),
                trans('report.6-12_hours'),
                trans('report.12-24_hours'),
                trans('report.1-2_days'),
                trans('report.2-4_days'),
                trans('report.4-7_days'),
                trans('report.7-14_days'),
                trans('report.14-30_days'),
                trans('report.30_plus_days'),
            ])
            ->setAxisNames(trans('general.count'), trans('general.time'))
            ->setOptions([
                'indexAxis' => 'y',
                'plugins' => [
                    'legend' => ['display' => false]
                ]
            ]);
    }

    public function underlyingDataTableColumns(): array
    {
        $timeColumn = (new Column(trans('report.time_to_resolution')))
            ->setId('time_to_resolution')
            ->setDefinition(['width' => '160px'])
            ->setValue(fn($ticket) => $this->formatTime((int) $ticket->time_to_resolution));

        return [
            $timeColumn,
            new TicketCreatedAtColumn,
            new TicketNumberColumn,
            new TicketSubjectColumn,
            new TicketDepartmentColumn,
        ];
    }

    public function underlyingDataQuery(Request $request): Builder
    {
        $query = $this->baseQuery()
            ->select(['ticket.*'])
            ->selectRaw('CAST(report_metric_value.value as UNSIGNED) as `time_to_resolution`');

        if ($this->isDrillIntoChartRequest($request)) {
            $label = $request->string('label')->toString();

            $exp = match ($label) {
                trans('report.0-6_hours') => $this->zeroTo6HoursExpr(),
                trans('report.6-12_hours') => $this->sixTo12HoursExpr(),
                trans('report.12-24_hours') => $this->twelveTo24HoursExpr(),
                trans('report.1-2_days') => $this->oneTo2DaysExpr(),
                trans('report.2-4_days') => $this->twoTo4DaysExpr(),
                trans('report.4-7_days') => $this->fourTo7DaysExpr(),
                trans('report.7-14_days') => $this->sevenTo14DaysExpr(),
                trans('report.14-30_days') => $this->fourteenTo30DaysExpr(),
                trans('report.30_plus_days') => $this->gt30DaysExpr(),
                default => throw new UnhandledMatchError
            };

            $query->whereRaw($exp);
        }

        return $query;
    }

    /**
     * @return int[]
     */
    private function getChartJsDataset(): array
    {
        $results = $this->baseQuery()
            ->selectRaw(sprintf('SUM(case when %s then 1 else 0 end) as `0`', $this->zeroTo6HoursExpr()))
            ->selectRaw(sprintf('SUM(case when %s then 1 else 0 end) as `1`', $this->sixTo12HoursExpr()))
            ->selectRaw(sprintf('SUM(case when %s then 1 else 0 end) as `2`', $this->twelveTo24HoursExpr()))
            ->selectRaw(sprintf('SUM(case when %s then 1 else 0 end) as `3`', $this->oneTo2DaysExpr()))
            ->selectRaw(sprintf('SUM(case when %s then 1 else 0 end) as `4`', $this->twoTo4DaysExpr()))
            ->selectRaw(sprintf('SUM(case when %s then 1 else 0 end) as `5`', $this->fourTo7DaysExpr()))
            ->selectRaw(sprintf('SUM(case when %s then 1 else 0 end) as `6`', $this->sevenTo14DaysExpr()))
            ->selectRaw(sprintf('SUM(case when %s then 1 else 0 end) as `7`', $this->fourteenTo30DaysExpr()))
            ->selectRaw(sprintf('SUM(case when %s then 1 else 0 end) as `8`', $this->gt30DaysExpr()))
            ->getQuery()
            ->get()
            ->firstOrFail();

        return array_values((array) $results);
    }

    private function zeroTo6HoursExpr(): string
    {
        return 'FLOOR(report_metric_value.value / 3600) < 6';
    }

    private function sixTo12HoursExpr(): string
    {
        return 'FLOOR(report_metric_value.value / 3600) >= 6 and FLOOR(report_metric_value.value / 3600) < 12';
    }

    private function twelveTo24HoursExpr(): string
    {
        return 'FLOOR(report_metric_value.value/ 3600) >= 12 and FLOOR(report_metric_value.value / 3600) < 24';
    }

    private function oneTo2DaysExpr(): string
    {
        return 'report_metric_value.value >= 86400 and report_metric_value.value < 172800';
    }

    private function twoTo4DaysExpr(): string
    {
        return 'report_metric_value.value >= 172800 and report_metric_value.value < 345600';
    }

    private function fourTo7DaysExpr(): string
    {
        return 'report_metric_value.value >= 345600 and report_metric_value.value < 604800';
    }

    private function sevenTo14DaysExpr(): string
    {
        return 'report_metric_value.value >= 604800 and report_metric_value.value < 1209600';
    }

    private function fourteenTo30DaysExpr(): string
    {
        return 'report_metric_value.value >= 1209600 and report_metric_value.value < 2592000';
    }

    private function gt30DaysExpr(): string
    {
        return 'report_metric_value.value >= 2592000';
    }

    /**
     * @return QueryBuilder<Ticket>
     */
    private function baseQuery(): QueryBuilder
    {
        return QueryBuilder::for(Ticket::class)
            ->applyFilters($this->dashboard()->filters('ticket.resolved_time'))
            ->join('report_metric_value', 'report_metric_value.model_id', '=', 'ticket.id')
            ->join('report_metric', 'report_metric_value.metric_id', '=', 'report_metric.id')
            ->where('ticket.status_id', Config::get('settings.default_resolved_status'))
            ->where('report_metric.name', '=', Metric::TIME_TO_RESOLUTION->value);
    }

    private function isDrillIntoChartRequest(Request $request): bool
    {
        return $request->has('label');
    }
}
