<?php declare(strict_types=1);

namespace Addons\Reports\TicketVolume\Reports\Cards\Overview;

use App\Http\Requests\Request;
use App\Modules\Report\Addon\Concerns\InteractsWithTimeframe;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\ChartType;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\Datasets\ObjectDataset;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\DataStructure;
use App\Modules\Report\Addon\Dashboards\Cards\ChartCard;
use App\Modules\Report\Addon\Dashboards\Cards\UnderlyingData;
use App\Modules\Report\Addon\Dashboards\Filtering\QueryBuilder;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketCreatedAtColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketNumberColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketReopenedTimeColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketResolvedTimeColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketStatusColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketSubjectColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\UserFormattedNameColumn;
use App\Modules\Ticket\Models\Ticket;
use Carbon\CarbonInterface;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Support\Collection;
use Illuminate\Support\ItemNotFoundException;
use UnhandledMatchError;

use function get_object_vars;
use function sprintf;
use function trans;
use function trans_choice;

/**
 * @implements UnderlyingData<Ticket>
 */
class TicketManagementChart extends ChartCard implements UnderlyingData
{
    use InteractsWithTimeframe;

    protected int $width = 12;

    public function name(): string
    {
        return trans('report.ticket_management');
    }

    public function data(): DataStructure
    {
        $unit = $this->calculateUnits();

        $dataset = new ObjectDataset($this->getOpenedDataset()->values()->toArray(), $this->openedLabel());
        $dataStructure = new DataStructure(ChartType::LINE, $dataset);

        $dataset = new ObjectDataset($this->getResolvedDataset()->values()->toArray(), $this->resolvedLabel());
        $dataStructure->addDataset($dataset);

        $dataset = new ObjectDataset($this->getReopenedDataset()->values()->toArray(), $this->reopenedLabel());

        return $dataStructure->addDataset($dataset)
            ->setAxisNames($this->unitToTrans($unit), trans_choice('ticket.ticket', 2))
            ->setOptions([
                'scales' => [
                    'x' => [
                        'type'    => 'time',
                        'time'    => [
                            'unit' => $unit,
                            'round' => $unit,
                            'tooltipFormat' => $this->unitToTooltipFormat($unit),
                        ],
                    ],
                    'y' => [
                        'min'   => 0,
                        'ticks' => ['precision' => 0],
                    ],
                ]
            ]);
    }

    /**
     * @inheritDoc
     */
    public function underlyingDataTableColumns(): array
    {
        return [
            new TicketCreatedAtColumn,
            new TicketNumberColumn,
            new TicketSubjectColumn,
            new UserFormattedNameColumn,
            new TicketStatusColumn,
            new TicketResolvedTimeColumn,
            new TicketReopenedTimeColumn,
        ];
    }

    /**
     * @inheritDoc
     */
    public function underlyingDataQuery(Request $request): Builder
    {
        if ($this->isDrillIntoChartRequest($request)) {
            $column = $this->labelToColumn($request->string('label')->toString());
            $query = $this->baseQuery($column);

            $time = $request->integer('value.x') / 1000;
            $query->whereRaw(sprintf('%s = ?', $this->getTimeSeriesExpr($column)), $time);
        } else {
            $query = $this->baseQuery('ticket.created_at')
                ->union($this->baseQuery('ticket.resolved_time')->select('ticket.*'))
                ->union($this->baseQuery('ticket.reopened_time')->select('ticket.*'));
        }

        return $query->with('user', 'status')
            ->select('ticket.*')
            ->orderBy('created_at', 'asc');
    }

    /**
     * @return mixed[]
     */
    public function underlyingDataTableOptions(): array
    {
        return ['ordering' => false];
    }

    /**
     * @return Builder<Ticket>
     */
    private function baseQuery(string $column): Builder
    {
        return QueryBuilder::for(Ticket::class)
            ->applyFilters($this->dashboard()->filters($column));
    }

    private function openedLabel(): string
    {
        return trans('general.opened');
    }

    private function resolvedLabel(): string
    {
        return trans('ticket.resolved');
    }

    private function reopenedLabel(): string
    {
        return trans('report.reopened');
    }

    /**
     * @return Collection<int,mixed>
     */
    private function getOpenedDataset(): Collection
    {
        return $this->getChartJsDataset('ticket.created_at');
    }

    /**
     * @return Collection<int,mixed>
     */
    private function getResolvedDataset(): Collection
    {
        return $this->getChartJsDataset('ticket.resolved_time');
    }

    /**
     * @return Collection<int,mixed>
     */
    private function getReopenedDataset(): Collection
    {
        return $this->getChartJsDataset('ticket.reopened_time');
    }

    /**
     * @return Collection<int,mixed>
     */
    private function getChartJsDataset(string $column): Collection
    {
        $collection = $this->baseQuery($column)
            ->selectRaw(sprintf('%s * 1000 as x', $this->getTimeSeriesExpr($column)))
            ->selectRaw('COUNT(*) as y')
            ->groupBy('x')
            ->getQuery()
            ->get()
            ->keyBy('x');

        return $this->createDateRange($this->calculateUnits())
            ->map(function (CarbonInterface $d) use ($collection) {
                if (isset($collection[$d->unix() * 1000])) {
                    return get_object_vars($collection[$d->unix() * 1000]);
                }

                return ['x' => $d->unix() * 1000, 'y' => 0];
            });
    }

    private function isDrillIntoChartRequest(Request $request): bool
    {
        return $request->has('label') && $request->has('value.x') && $request->has('value.y');
    }

    /**
     * @throws UnhandledMatchError|ItemNotFoundException
     */
    private function labelToColumn(string $label): string
    {
        return match ($label) {
            $this->openedLabel()   => 'ticket.created_at',
            $this->resolvedLabel() => 'ticket.resolved_time',
            $this->reopenedLabel() => 'ticket.reopened_time',
            default                =>  throw new UnhandledMatchError(sprintf('Unsupported label \'%s\'.', $label)),
        };
    }
}
