<?php declare(strict_types=1);

namespace Addons\Reports\TicketVolume\Reports\Cards\Workload;

use App\Http\Requests\Request;
use App\Modules\Report\Addon\Dashboards\Cards\Table\Column;
use App\Modules\Report\Addon\Dashboards\Cards\TableCard;
use App\Modules\Report\Addon\Dashboards\Cards\UnderlyingData;
use App\Modules\Report\Addon\Dashboards\Filtering\QueryBuilder;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketCreatedAtColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketDepartmentColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketNumberColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketResolvedTimeColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketSubjectColumn;
use App\Modules\Ticket\Models\Ticket;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Support\Facades\Config;

use function number_format;
use function trans;
use function trans_choice;

/**
 * @phpstan-extends TableCard<Ticket>
 * @implements UnderlyingData<Ticket>
 */
class AverageMessagesToResolutionTable extends TableCard implements UnderlyingData
{
    protected int $width = 6;

    public function name(): string
    {
        return trans('report.average_messages_to_resolution_per_department');
    }

    /**
     * @inheritDoc
     */
    public function query(): QueryBuilder
    {
        return $this->baseQuery()
            ->select('department.name', 'ticket.department_id')
            ->selectRaw('AVG(ticket.messages_count) as average')
            ->groupBy('ticket.department_id');
    }

    public function columns(): array
    {
        return [
            $this->departmentColumn(),
            $this->averageColumn()
        ];
    }

    /**
     * @inheritDoc
     */
    public function underlyingDataQuery(Request $request): Builder
    {
        $query = $this->baseQuery();

        if ($this->isDrillIntoChartRequest($request)) {
            $query->where('ticket.department_id', $request->integer('value'));
        }

        return $query->with('user', 'department')
            ->selectRaw('ticket.*');
    }

    /**
     * @inheritDoc
     */
    public function underlyingDataTableColumns(): array
    {
        return [
            new TicketCreatedAtColumn,
            new TicketNumberColumn,
            new TicketSubjectColumn,
            new TicketDepartmentColumn,
            new TicketResolvedTimeColumn,
            $this->noOfMessagesColumn()
        ];
    }

    public function getUnderlyingDataValueKey(): string
    {
        return 'department_id';
    }

    /**
     * @return QueryBuilder<Ticket>
     */
    private function baseQuery(): QueryBuilder
    {
        return QueryBuilder::for(Ticket::class)
            ->applyFilters($this->dashboard()->filters('ticket.created_at'))
            ->join('department', 'ticket.department_id', '=', 'department.id')
            ->where('ticket.status_id', Config::get('settings.default_resolved_status'))
            ->whereNotNull('ticket.messages_count');
    }

    private function departmentColumn(): Column
    {
        return (new Column(trans_choice('ticket.department', 1)))
            ->setId('name')
            ->setValue(fn($model) => $model->name);
    }

    private function averageColumn(): Column
    {
        return (new Column(trans('report.average_messages_to_resolution')))
            ->setId('average')
            ->setValue(fn($model) => number_format((float) $model->average, 2));
    }

    private function noOfMessagesColumn(): Column
    {
        return (new Column(trans('report.no_of_messages')))
            ->setId('messages_count')
            ->setValue(fn($model) => $model->messages_count);
    }

    private function isDrillIntoChartRequest(Request $request): bool
    {
        return $request->has('value');
    }
}
