<?php declare(strict_types=1);

namespace SupportPal\TopoClient;

use ArrayAccess;
use Illuminate\Contracts\Support\Arrayable;
use InvalidArgumentException;
use RuntimeException;

use function array_filter;
use function array_key_exists;

/**
 * @phpstan-import-type Topology from Geometry
 * @phpstan-type Properties array{'name': string}
 *
 * @implements Arrayable<string, mixed>
 * @implements ArrayAccess<string, mixed>
 */
class Feature implements Arrayable, ArrayAccess
{
    private const TYPE = 'Feature';
    private ?string $id = null;
    /** @var float[]|null */
    private ?array $bbox = null;
    /** @var Properties */
    private readonly array $properties;
    /** @var mixed[] */
    private readonly array $geometry;

    /**
     * @param Properties $properties
     * @param mixed[] $geometry
     */
    public function __construct(array $properties, array $geometry)
    {
        $this->properties = $properties;
        $this->geometry = $geometry;
    }

    public function getId(): ?string
    {
        return $this->id;
    }

    public function setId(string $id): self
    {
        $this->id = $id;

        return $this;
    }

    /**
     * @param float[] $bbox
     */
    public function setBbox(array $bbox): self
    {
        $this->bbox = $bbox;

        return $this;
    }

    public function toArray(): array
    {
        return array_filter([
            'type' => self::TYPE,
            'id'   => $this->id,
            'bbox' => $this->bbox,
            'properties' => $this->properties,
            'geometry' => $this->geometry,
        ]);
    }

    /**
     * @param Topology $topology
     * @param mixed[] $o
     */
    public static function parse(array $topology, array $o): self
    {
        $properties = array_key_exists('properties', $o) ? $o['properties'] : [];
        $geometry   = (new Geometry($topology))->parse($o)->toArray();
        $instance = new self($properties, $geometry);
        if (isset($o['id'])) {
            $instance->setId($o['id']);
        }

        if (isset($o['bbox'])) {
            $instance->setBbox($o['bbox']);
        }

        return $instance;
    }

    public function offsetExists(mixed $offset): bool
    {
        return array_key_exists($offset, $this->toArray());
    }

    public function offsetGet(mixed $offset): mixed
    {
        return $this->toArray()[$offset] ?? throw new InvalidArgumentException('Undefined offset.');
    }

    public function offsetSet(mixed $offset, mixed $value): void
    {
        throw new RuntimeException('The offset is readonly.');
    }

    public function offsetUnset(mixed $offset): void
    {
        throw new RuntimeException('The offset is readonly.');
    }
}
