<?php declare(strict_types=1);

namespace SupportPal\TopoClient;

use Closure;
use SupportPal\TopoClient\Geometry\Collection;
use SupportPal\TopoClient\Geometry\Coordinates;
use SupportPal\TopoClient\Geometry\Type;
use UnhandledMatchError;

use function array_map;
use function array_pop;
use function call_user_func;
use function count;

/**
 * @phpstan-type Topology array{'type': 'Topology', 'objects': mixed[], 'arcs': Arcs, 'transform': Transform}
 * @phpstan-type Arcs array<int, array<int, int[]>>
 * @phpstan-type Transform array{'scale': float[], 'translate': int[]}
 */
class Geometry
{
    private Closure $transformPoint;
    /** @var Arcs */
    private array $arcs;
    private int $x0;
    private int $y0;

    /**
     * @param Topology $topology
     */
    public function __construct(array $topology)
    {
        $this->transformPoint = $this->transform($topology['transform']);
        $this->arcs = $topology['arcs'];
        $this->x0 = 0;
        $this->y0 = 0;
    }

    /**
     * @phpstan-param array<mixed, mixed> $o
     */
    public function parse(array $o): Collection|Coordinates
    {
        $type = $o['type'];
        if ($type === 'GeometryCollection') {
            return new Collection(array_map(fn($geometry) => $this->parse($geometry), $o['geometries']));
        }

        return match ($type) {
            'Point' => new Coordinates(
                Type::POINT,
                $this->point($o['coordinates'])
            ),
            'MultiPoint' => new Coordinates(
                Type::MultiPoint,
                array_map(fn($coordinate) => $this->point($coordinate), $o['coordinates'])
            ),
            'LineString' => new Coordinates(
                Type::LineString,
                $this->line($o['arcs'])
            ),
            'MultiLineString' => new Coordinates(
                Type::MultiLineString,
                array_map(fn($arc) => $this->line($arc), $o['arcs'])
            ),
            'Polygon' => new Coordinates(
                Type::Polygon,
                $this->polygon($o['arcs'])
            ),
            'MultiPolygon' => new Coordinates(
                Type::MultiPolygon,
                array_map(fn($arc) => $this->polygon($arc), $o['arcs'])
            ),
            default => throw new UnhandledMatchError
        };
    }

    /**
     * @param float[] $points
     * @return float[]
     */
    private function arc(int $i, array $points): array
    {
        if (count($points)) {
            array_pop($points);
        }

        for ($a = $this->arcs[$i < 0 ? ~$i : $i], $k = 0, $n = count($a); $k < $n; ++$k) {
            $points[] = call_user_func($this->transformPoint, $a[$k], $k);
        }

        if ($i >= 0) {
            return $points;
        }

        return $this->reverse($points, $n);
    }

    /**
     * @param mixed[]|null $p
     * @return float[]
     */
    private function point(?array $p): array
    {
        return call_user_func($this->transformPoint, $p);
    }

    /**
     * @param int[] $arcs
     * @return float[]
     */
    private function line(array $arcs): array
    {
        $points = [];
        for ($i = 0, $n = count($arcs); $i < $n; ++$i) {
            $points = $this->arc($arcs[$i], $points);
        }

        // This should never happen per the specification.
        if (count($points) < 2) {
            $points[] = $points[0];
        }

        return $points;
    }

    /**
     * @param int[] $arcs
     * @return float[]
     */
    private function ring(array $arcs): array
    {
        $points = $this->line($arcs);

        // This may happen if an arc has only two points.
        while (count($points) < 4) {
            $points[] = $points[0];
        }

        return $points;
    }

    /**
     * @param array<int, int[]> $arcs
     * @return array<int, float[]>
     */
    private function polygon(array $arcs): array
    {
        return array_map(fn (array $arc) => $this->ring($arc), $arcs);
    }

    /**
     * @param Transform|null $transform
     */
    private function transform(?array $transform): Closure
    {
        if ($transform === null) {
            return function ($x) {
                return $x;
            };
        }

        $kx = $transform['scale'][0];
        $ky = $transform['scale'][1];
        $dx = $transform['translate'][0];
        $dy = $transform['translate'][1];

        return function ($input, $i) use ($kx, $ky, $dx, $dy) {
            if (! $i) {
                $this->x0 = $this->y0 = 0;
            }

            $j = 2;
            $n = count($input);
            $output = [$n];
            $output[0] = ($this->x0 += $input[0]) * $kx + $dx;
            $output[1] = ($this->y0 += $input[1]) * $ky + $dy;
            while ($j < $n) {
                $output[$j] = $input[$j];
                ++$j;
            }

            return $output;
        };
    }

    /**
     * @param float[] $array
     * @return float[]
     */
    private function reverse(array $array, int $n): array
    {
        $j = count($array);
        $i = $j - $n;
        while ($i < --$j) {
            $t = $array[$i];
            $array[$i++] = $array[$j];
            $array[$j] = $t;
        }

        return $array;
    }
}
