<?php declare(strict_types=1);

namespace Database\Seeds\v5_0_0;

use App\Exceptions\Handler;
use App\Modules\Core\Controllers\Addons\Manager\AddonManager;
use App\Modules\Core\Controllers\Database\Seed\Seeder;
use Exception;
use Illuminate\Contracts\Container\Container;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\DB;
use PluginFactory;
use SupportPal\Addons\Addon;
use SupportPal\Marketplace\Http\Api\Addons;

use function collect;
use function count;
use function sprintf;

class SeedPluginTable extends Seeder
{
    private Addons $addonsApi;

    private AddonManager $addonManager;

    public function __construct(Container $container, Addons $addonsApi, AddonManager $addonManager)
    {
        parent::__construct($container);

        $this->addonsApi = $addonsApi;
        $this->addonManager = $addonManager;
    }

    /**
     * Run the database seeds.
     *
     * @return void
     */
    public function run()
    {
        $corePlugins = ['BlestaInformation', 'GoogleTranslate', 'SlackNotifications', 'TimeTracking', 'WhmcsInformation', 'WooCommerce'];

        $plugins = DB::table('plugin')->whereIn('name', $corePlugins)->where('enabled', 1)->get();

        $this->deactivatePlugins($plugins);

        foreach ($plugins as $plugin) {
            $addons = $this->addonsApi->get(['categories' => ['plugins' => [$plugin->name]]]);

            if ($addons->isEmpty()) {
                $this->note(sprintf('[ERROR] Unable to find plugin %s at marketplace.', $plugin->name));

                continue;
            }

            try {
                $addon = $addons->first();
                $this->addonManager->update($addon);
                $this->note(sprintf('[NOTICE] Updated plugin %s from marketplace.', $plugin->name));
            } catch (Exception $e) {
                if ($this->container->runningUnitTests()) {
                    throw $e;
                }

                Handler::reportException($e);
                $this->note(sprintf('[NOTICE] Failed to update plugin %s from marketplace.', $plugin->name));
            }
        }
    }

    /**
     * Deactivate enabled plugins.
     *
     * @param Collection<int,mixed[]> $corePlugins
     */
    private function deactivatePlugins(Collection $corePlugins): void
    {
        $plugins = DB::table('plugin')
            ->where('enabled', 1)
            ->get();

        foreach ($plugins as $id => $row) {
            $plugin = PluginFactory::getPlugin($row->name);

            if (! $plugin instanceof Addon) {
                continue;
            }

            try {
                PluginFactory::deactivate($plugin, true);
            } catch (Exception $e) {
                $this->note(sprintf("[ERROR] Failed to deactivate '%s' plugin. Reason: '%s'.", $row->name, $e->getMessage()));
                Handler::reportException($e);

                continue;
            }

            // Only keep track of deactivated plugins that are not included in core.
            if (! $corePlugins->contains('name', $row->name)) {
                continue;
            }

            $plugins->forget($id);
        }

        if (count($plugins) <= 0) {
            return;
        }

        $this->note('[NOTICE] The following plugins have been deactivated due to potential compatibility issues'
            . ' with this release: ' . collect($plugins)->implode('name', ', ') . '. Please see'
            . ' https://docs.supportpal.com/current/Upgrading+to+5.0+from+4.2#DevelopmentChanges for more information.');
    }
}
