<?php declare(strict_types=1);

namespace SupportPal\BinPacking;

use Illuminate\Support\Collection;

use function count;
use function uasort;

class BinPacking
{
    protected int $maximumBinWeight;

    public function __construct(int $maximumBinWeight)
    {
        $this->maximumBinWeight = $maximumBinWeight;
    }

    /**
     * Packs a list of item into bins.
     *
     * @template TKey of array-key
     * @param array<TKey, Item> $items
     * @return Collection<TKey, Bin>
     */
    public function pack(array $items): Collection
    {
        // Sort in descending order by weight
        uasort($items, function ($item1, $item2) {
            return $item2->getWeight() - $item1->getWeight();
        });

        // We will use, at most, the same number of items in bins.
        // We will filter out the empty bins at the end
        /** @var Collection<TKey, Bin> $bins */
        $bins = new Collection;

        for ($i = 0; $i < count($items); $i++) {
            $bins[] = new Bin($this->maximumBinWeight);
        }

        foreach ($items as $key => $item) {
            foreach ($bins as $bin) {
                if ($bin->itemFits($item)) {
                    $bin->addItem($key, $item);
                    break;
                }
            }
        }

        // Remove empty bins
        return $bins->filter(fn(Bin $bin) => $bin->getTotalWeight() > 0);
    }
}
