<?php declare(strict_types=1);

namespace Chumper\Datatable\Database;

use Illuminate\Contracts\Database\Query\Builder;
use Illuminate\Database\Connection;
use Illuminate\Support\Str;

use function sprintf;

class Counter
{
    /**
     * Flag to ignore the selects in count query.
     *
     * @var bool
     */
    protected bool $ignoreSelectInCountQuery = false;

    /**
     * Flag to keep the select bindings.
     *
     * @var bool
     */
    protected bool $keepSelectBindings = false;

    public function __construct(protected Builder $query)
    {
        //
    }

    /**
     * Ignore the selects in count query.
     */
    public function ignoreSelectsInCountQuery(bool $value): static
    {
        $this->ignoreSelectInCountQuery = $value;

        return $this;
    }

    /**
     * Keep the select bindings.
     */
    public function keepSelectBindings(bool $value): static
    {
        $this->keepSelectBindings = $value;

        return $this;
    }

    public function count(): int
    {
        return $this->prepareCountQuery()->count();
    }

    private function prepareCountQuery(): Builder
    {
        $builder = clone $this->query;

        if ($this->isComplexQuery($builder)) {
            $builder->selectRaw('1 as dt_row_count');
            if ($this->ignoreSelectInCountQuery || ! $this->isComplexQuery($builder)) {
                return $this->getConnection()
                    ->query()
                    ->fromRaw('('.$builder->toSql().') count_row_table')
                    ->setBindings($builder->getBindings());
            }

            $builder = clone $this->query;

            return $this->getConnection()
                ->query()
                ->fromRaw('('.$builder->toSql().') count_row_table')
                ->setBindings($builder->getBindings());
        }

        $row_count = $this->wrap('row_count');
        $builder->select($this->getConnection()->raw(sprintf('\'1\' as %s', $row_count)));
        if (! $this->keepSelectBindings) {
            $builder->setBindings([], 'select');
        }

        return $builder;
    }

    /**
     * Wrap column with DB grammar.
     *
     * @param  string  $column
     * @return string
     */
    protected function wrap(string $column): string
    {
        return $this->getConnection()->getQueryGrammar()->wrap($column);
    }

    protected function isComplexQuery(Builder $query): bool
    {
        return Str::contains(Str::lower($query->toSql()), ['union', 'having', 'distinct', 'order by', 'group by']);
    }

    private function getConnection(): Connection
    {
        /** @var Connection $connection */
        $connection = $this->query->getConnection();

        return $connection;
    }
}
