<?php declare(strict_types=1);

namespace Chumper\Datatable\Engines;

use Chumper\Datatable\Engines\Dataset\CollectionDataset;
use Chumper\Datatable\Engines\Dataset\Dataset;
use Illuminate\Http\Request;
use Illuminate\Support\Collection;
use Illuminate\Support\Str;

use function array_diff_key;
use function array_filter;
use function array_key_exists;
use function array_search;
use function count;
use function in_array;
use function is_callable;
use function mb_strtolower;
use function strip_tags;

use const SORT_NATURAL;

/**
 * This handles the collections,
 * it needs to compile first, so we wait for the make command and then
 * do all the operations
 *
 * Class CollectionEngine
 * @package Chumper\Datatable\Engines
 */
class CollectionEngine extends BaseEngine
{
    /**
     * Constant for OR queries in internal search
     *
     * @var string
     */
    const OR_CONDITION = 'OR';
    /**
     * Constant for AND queries in internal search
     *
     * @var string
     */
    const AND_CONDITION = 'AND';

    /** @var Collection */
    private $workingCollection;

    /** @var Collection */
    private $collection;

    /** @var array Different options */
    private $options = array(
        'sortFlags'         => SORT_NATURAL,
        'stripOrder'        => false,
        'stripSearch'       => false,
        'caseSensitive'     => false,
    );

    public function __construct(Collection $collection, Request $request)
    {
        parent::__construct($request);
        $this->collection = $collection;
        $this->workingCollection = $collection;
    }

    public function count(): int
    {
        return $this->workingCollection->count();
    }

    public function totalCount(): int
    {
        return $this->collection->count();
    }

    public function getArray(): array
    {
        return $this->dataset()->get()->toArray();
    }

    /**
     * Resets all operations performed on the collection
     */
    public function reset(): static
    {
        $this->workingCollection = $this->collection;

        return $this;
    }

    public function stripSearch(): static
    {
        $this->options['stripSearch'] = true;

        return $this;
    }

    /**
     * @param bool|callable $callback
     */
    public function stripOrder($callback = true): static
    {
        $this->options['stripOrder'] = $callback;

        return $this;
    }

    public function setSearchStrip(): static
    {
        $this->options['stripSearch'] = true;

        return $this;
    }

    /**
     * @param bool|callable $callback
     */
    public function setOrderStrip($callback = true): static
    {
        return $this->stripOrder($callback);
    }

    /**
     * Set the sort behaviour of the doInternalOrder() function.
     *
     * @param int $sort_flags For details see: http://php.net/manual/en/function.sort.php
     */
    public function setOrderFlags($sort_flags = SORT_NATURAL): static
    {
        $this->options['sortFlags'] = $sort_flags;

        return $this;
    }

    /**
     * @param bool $value
     */
    public function setCaseSensitive($value): static
    {
        $this->options['caseSensitive'] = $value;

        return $this;
    }

    /**
     * @param string $value
     */
    public function getOption($value): mixed
    {
        return $this->options[$value];
    }

    protected function handleiDisplayLength(mixed $value): void
    {
        // Setting the value to -1 affects array_slice and removes the last elements
        // from the collection. -1 is effectively the same as null when used with array_slice
        if ((int) $value === -1) {
            $value = null;
        }

        parent::handleiDisplayLength($value);
    }

    protected function configureDataset(): CollectionDataset
    {
        $dataset = new CollectionDataset($this->workingCollection, $this);
        $dataset->setPagination($this->skip, $this->limit);

        return $dataset;
    }

    protected function countRecords(): int
    {
        return $this->totalCount();
    }

    /**
     * Filter a collection based on the DataTables search parameters (sSearch_0 etc)
     * See http://legacy.datatables.net/usage/server-side
     */
    protected function filter(Dataset $dataset): Dataset
    {
        if (empty($this->search) && empty($this->fieldSearches)) {
            return $dataset;
        }

        $value = $this->search;
        $caseSensitive = $this->options['caseSensitive'];

        $toSearch = array();
        $searchType = self::AND_CONDITION;
        // Map the searchColumns to the real columns
        $ii = 0;
        foreach ($this->columns as $col) {
            if (in_array($col->getName(), $this->searchColumns) || in_array($col->getName(), $this->fieldSearches)) {
                // map values to columns, where there is no value use the global value
                if (($field = array_search($col->getName(), $this->fieldSearches)) !== false) {
                    $toSearch[$ii] = $this->columnSearches[$field];
                } else {
                    if ($value) {
                        $searchType = self::OR_CONDITION;
                    }

                    $toSearch[$ii] = $value;
                }
            }

            $ii++;
        }

        $self = $this;
        $this->workingCollection = $dataset->raw()->filter(function ($row) use ($toSearch, $caseSensitive, $self, $searchType) {
            for ($i = 0, $stack = array(), $nb = count($row); $i < $nb; $i++) {
                if (! array_key_exists($i, $toSearch)) {
                    continue;
                }

                $column = $i;
                if ($self->getAliasMapping()) {
                    $column = $self->getNameByIndex($i);
                }

                if ($self->getOption('stripSearch')) {
                    $search = strip_tags($row[$column]);
                } else {
                    $search = $row[$column];
                }

                // sburkett - added support for exact matching on specific columns
                if (@$this->columnSearchExact[ $self->getNameByIndex($i) ] === true) {
                    if ($toSearch[$i] === $search) {
                        $stack[$i] = true;
                    }
                } else {
                    if ($caseSensitive) {
                        if ($self->exactWordSearch) {
                            if ($toSearch[$i] === $search) {
                                $stack[$i] = true;
                            }
                        } else {
                            if (Str::contains($search, $toSearch[$i])) {
                                $stack[$i] = true;
                            }
                        }
                    } else {
                        if ($self->getExactWordSearch()) {
                            if (mb_strtolower($toSearch[$i]) === mb_strtolower($search)) {
                                $stack[$i] = true;
                            }
                        } else {
                            if (Str::contains(mb_strtolower($search), mb_strtolower($toSearch[$i]))) {
                                $stack[$i] = true;
                            }
                        }
                    }
                }
            }

            if ($searchType === $self::AND_CONDITION) {
                $result = array_diff_key(array_filter($toSearch), $stack);

                return empty($result);
            }

            return ! empty($stack);
        });

        return $dataset->set($this->workingCollection);
    }

    protected function sort(Dataset $dataset): Dataset
    {
        if ($this->orderColumn === null) {
            return $dataset;
        }

        $column = $this->orderColumn[0];
        $stripOrder = $this->options['stripOrder'];
        $self = $this;
        $this->workingCollection = $dataset->raw()->sortBy(function ($row) use ($column, $stripOrder, $self) {
            if ($self->getAliasMapping()) {
                $column = $self->getNameByIndex($column);
            }

            if ($stripOrder) {
                if (is_callable($stripOrder)) {
                    return $stripOrder($row, $column);
                }

                return strip_tags($row[$column]);
            }

            return $row[$column];
        }, $this->options['sortFlags']);

        if ($this->orderDirection === BaseEngine::ORDER_DESC) {
            $this->workingCollection = $this->workingCollection->reverse();
        }

        return $dataset->set($this->workingCollection);
    }
}
