<?php declare(strict_types=1);

namespace Chumper\Datatable;

use Exception;
use Illuminate\Support\Facades\Config;
use Illuminate\Support\Facades\Request;
use Illuminate\Support\Facades\View;
use Illuminate\Support\Str;

use function array_combine;
use function array_key_exists;
use function array_merge;
use function array_merge_recursive;
use function call_user_func;
use function count;
use function is_array;
use function is_callable;
use function is_numeric;
use function is_string;
use function json_encode;
use function preg_match;
use function sprintf;
use function trim;

/**
 * Class Table
 * @package Chumper\Datatable
 */
class Table
{
    /** @var array */
    private $config = array();

    /** @var array */
    private $columns = array();

    /** @var array */
    private $options = array();

    /** @var array */
    private $callbacks = array();

    /**
     * Values to be sent to custom templates
     *
     * @var array
     */
    private $customValues = array();

    /** @var array */
    private $data = array();

    /** @var boolean Determines if the template should echo the javascript */
    private $noScript = false;

    /** @var String The name of the id the table will have later */
    protected $idName;

    /** @var String The name of the class the table will have later */
    protected $className;

    /** @var String The view used to render the table */
    protected $table_view;

    /** @var String The view used to render the javascript */
    protected $script_view;

    /** @var boolean indicates if the mapping was already added to the options */
    private $createdMapping = true;

    /** @var array name of mapped columns */
    private $aliasColumns = array();

    public function __construct()
    {
        $this->config = Config::get('chumper.datatable.table');

        $this->setId($this->config['id']);
        $this->setClass($this->config['class']);
        $this->setOptions($this->config['options']);
        $this->setCallbacks($this->config['callbacks']);

        $this->noScript = $this->config['noScript'];
        $this->table_view = $this->config['table_view'];
        $this->script_view = $this->config['script_view'];
    }

    public function addColumn(string ...$columns): self
    {
        foreach ($columns as $title) {
            $this->columns[] = $title;
            $this->aliasColumns[] = count($this->aliasColumns) + 1;
        }

        return $this;
    }

    public function countColumns(): int
    {
        return count($this->columns);
    }

    /**
     * @param string $key
     */
    public function removeOption($key): self
    {
        if (isset($this->options[$key])) {
            unset($this->options[$key]);
        }

        return $this;
    }

    /**
     * Set JavaScript config definition for DataTables.
     *
     * @param array|string $options
     */
    public function setOptions(string|array $options, string $value = ''): self
    {
        if (is_string($options)) {
            $options = [$options => $value];
        }

        foreach ($options as $key => $option) {
            $this->options[$key] = $this->getValue($option);
        }

        return $this;
    }

    /**
     * @param array $order
     */
    public function setOrder($order = []): self
    {
        $_orders = array();
        foreach ($order as $number => $sort) {
            $_orders[] = [$number, $sort];
        }

        $this->callbacks['aaSorting'] = $_orders;

        return $this;
    }

    public function setCallbacks(string|array $callbacks, string $value = ''): self
    {
        if (is_string($callbacks)) {
            $callbacks = [$callbacks => $value];
        }

        foreach ($callbacks as $key => $value) {
            $this->callbacks[$key] = $value;
        }

        return $this;
    }

    /**
     * @return $this
     * @throws Exception
     */
    public function setCustomValues(array $values): self
    {
        foreach ($values as $key => $value) {
            $this->customValues[$key] = $value;
        }

        return $this;
    }

    /**
     * @param array $data
     * @return $this
     */
    public function setData(array $data)
    {
        $this->data = $data;

        return $this;
    }

    public function setUrl(string $url): self
    {
        $this->options['sAjaxSource'] = $url;
        $this->options['bServerSide'] = true;

        return $this;
    }

    /**
     * @return array
     */
    public function getOptions()
    {
        return $this->options;
    }

    /**
     * @return array
     */
    public function getCallbacks()
    {
        return $this->callbacks;
    }

    /**
     * @return array
     */
    public function getCustomValues()
    {
        return $this->customValues;
    }

    /**
     * @return array
     */
    public function getData()
    {
        return $this->data;
    }

    public function render(?string $view = null): \Illuminate\Contracts\View\View
    {
        if ($view !== null) {
            $this->table_view = $view;
        }

        return View::make($this->table_view, $this->getViewParameters());
    }

    /**
     * returns an array with the parameters that will be passed to the view when it's rendered
     * @return array
     */
    public function getViewParameters()
    {
        if (! isset($this->options['sAjaxSource'])) {
            $this->setUrl(Request::url());
        }

        // create mapping for frontend
        if (! $this->createdMapping) {
            $this->createMapping();
        }

        return array(
            'options'   => $this->getOptionsAsJson(),
            'values'    => $this->customValues,
            'data'      => $this->data,
            'columns'   => array_combine($this->aliasColumns, $this->columns),
            'noScript'  => $this->noScript,
            'id'        => $this->idName,
            'class'     => $this->className,
        );
    }

    /**
     * Instructs the table not to echo the javascript
     *
     * @return $this
     */
    public function noScript()
    {
        $this->noScript = true;

        return $this;
    }

    private function convertData(array $options)
    {
        $is_obj = false;
        $first = true;
        $data = '';
        foreach ($options as $k => $o) {
            if ($first === true) {
                if (! is_numeric($k)) {
                    $is_obj = true;
                }

                $first = false;
            } else {
                $data .= ",\n";
            }

            if (! is_numeric($k)) {
                $data .= json_encode($k) . ':';
            }

            if (is_string($o)) {
                if (@preg_match('#^\s*function\s*\([^\)]*#', $o)) {
                    $data .= $o;
                } else {
                    $data .= json_encode($o);
                }
            } else {
                if (is_array($o)) {
                    $data .= $this->convertData($o);
                } else {
                    $data .= json_encode($o);
                }
            }
        }

        if ($is_obj) {
            return sprintf('{ %s }', $data);
        }

        return sprintf('[ %s ]', $data);
    }

    public function script(?string $view = null): \Illuminate\Contracts\View\View
    {
        if ($view !== null) {
            $this->script_view = $view;
        }

        // create mapping for frontend
        if (! $this->createdMapping) {
            $this->createMapping();
        }

        return View::make($this->script_view, array(
            'options' => $this->getOptionsAsJson(),
            'id'      => $this->idName,
        ));
    }

    public function getOptionsAsJson(): string
    {
        return $this->convertData(array_merge($this->options, $this->callbacks));
    }

    public function getId()
    {
        return $this->idName;
    }

    public function setId(string $id = ''): self
    {
        $this->idName = empty($id) ? Str::random(8) : $id;

        return $this;
    }

    public function getClass()
    {
        return $this->className;
    }

    public function setClass(string $class): self
    {
        $this->className = trim(sprintf('%s %s', $this->className, $class));

        return $this;
    }

    public function setAliasMapping(bool $value): self
    {
        $this->createdMapping = ! $value;

        return $this;
    }

    //--------------------PRIVATE FUNCTIONS

    private function createMapping()
    {
        // set options for better handling
        // merge with existing options
        if (! array_key_exists('aoColumns', $this->options)) {
            $this->options['aoColumns'] = array();
        }

        $matching = array();
        $i = 0;
        foreach ($this->aliasColumns as $name) {
            if (array_key_exists($i, $this->options['aoColumns'])) {
                $this->options['aoColumns'][$i] = array_merge_recursive($this->options['aoColumns'][$i], array('mData' => $name));
            } else {
                $this->options['aoColumns'][$i] = array('mData' => $name);
            }

            $i++;
        }

        $this->createdMapping = true;

        return $matching;
    }

    /**
     * If a variable is callable, get the return value.
     *
     * @param  mixed $variable
     * @return mixed
     */
    private function getValue($variable)
    {
        if (is_callable($variable)) {
            return call_user_func($variable);
        }

        return $variable;
    }
}
