<?php declare(strict_types=1);

namespace SupportPal\HealthService\Service;

use ArrayObject;
use Generator;
use Illuminate\Config\Repository as ConfigRepository;
use Illuminate\Contracts\Cache\Repository as CacheRepository;
use SupportPal\HealthService\Check\HealthCheck;
use SupportPal\HealthService\Model\Result;
use SupportPal\HealthService\Model\Translation;

use function get_class;

class HealthChecker
{
    /** @var array<string, HealthCheck> */
    private array $healthChecks;

    private CacheRepository $cacheManager;

    /** @var int|null */
    private $cacheTtl;

    /**
     * @param HealthCheck[] $healthChecks
     */
    public function __construct(array $healthChecks, CacheRepository $cacheManager, ConfigRepository $configRepository)
    {
        $this->healthChecks = [];
        foreach ($healthChecks as $healthCheck) {
            $this->healthChecks[get_class($healthCheck)] = $healthCheck;
        }

        $this->cacheManager = $cacheManager;
        $this->cacheTtl = $configRepository->get('cache.health.ttl');
    }

    /**
     * @return Generator<Result>
     */
    public function runChecks(bool $allowCache = false): Generator
    {
        $arrayCache = $allowCache ? $this->getCachedData() : new ArrayObject;

        foreach ($this->healthChecks as $className => $healthCheck) {
            if (! $this->checkDependencies($healthCheck, $arrayCache)) {
                yield Result::error($healthCheck->name(), new Translation('core.dependency_resolution'));

                continue;
            }

            if (! isset($arrayCache[$className])) {
                $arrayCache[$className] = $healthCheck->check();
            }

            yield $arrayCache[$className];
        }

        /**
         * @var string $check
         * @var Result $result
         */
        foreach ($arrayCache as $check => $result) {
            $this->cacheManager->add($check, $result, $this->cacheTtl);
        }
    }

    private function getCachedData(): ArrayObject
    {
        $cache = new ArrayObject;
        foreach ($this->healthChecks as $healthCheck) {
            $healthCheckClass = get_class($healthCheck);
            $cachedResult = $this->cacheManager->get($healthCheckClass);
            if ($cachedResult === null) {
                continue;
            }

            $cache[$healthCheckClass] = $cachedResult;
        }

        return $cache;
    }

    public function purgeCache(): void
    {
        foreach ($this->healthChecks as $className => $healthCheck) {
            $this->cacheManager->forget($className);
        }
    }

    /**
     * @param HealthCheck $healthCheck
     * @param ArrayObject $arrayCache
     * @return bool
     */
    private function checkDependencies(HealthCheck $healthCheck, ArrayObject $arrayCache): bool
    {
        foreach ($healthCheck->dependsOn() as $checkDependencyClass) {
            if (! isset($arrayCache[$checkDependencyClass])) {
                $arrayCache[$checkDependencyClass] = $this->healthChecks[$checkDependencyClass]->check();
            }

            $result = $arrayCache[$checkDependencyClass];
            if ($result->getState() === Result::ERROR) {
                return false;
            }
        }

        return true;
    }
}
