<?php declare(strict_types=1);

namespace SupportPal\Marketplace\Http;

use GuzzleHttp\Client;
use GuzzleHttp\Exception\BadResponseException;
use GuzzleHttp\Exception\TransferException;
use GuzzleHttp\HandlerStack;
use GuzzleHttp\Middleware;
use Illuminate\Support\Facades\Cache;
use Illuminate\Support\Facades\Config;
use Kevinrob\GuzzleCache\CacheMiddleware;
use Kevinrob\GuzzleCache\Storage\LaravelCacheStorage;
use Kevinrob\GuzzleCache\Strategy\PublicCacheStrategy;
use Psr\Log\LoggerInterface;
use SupportPal\Marketplace\Exceptions\ApiException;
use SupportPal\ProxyService\ProxyService;

use function env;
use function fclose;
use function fopen;
use function json_decode;

class Request
{
    private ProxyService $proxyService;
    private LoggerInterface $logger;

    public function __construct(ProxyService $proxyService, LoggerInterface $logger)
    {
        $this->proxyService = $proxyService;
        $this->logger = $logger;
    }

    /**
     * Domains where we can access the marketplace Api.
     * Trailing slash is required for Guzzle 6 base_uri setting to work with directory paths.
     */
    public function getBaseUrl(): string
    {
        return env('MARKETPLACE_URL', 'https://marketplace.supportpal.com') . '/api/';
    }

    /**
     * Perform a GET http request.
     *
     * @param  mixed[]  $parameters
     * @return mixed
     * @throws ApiException
     */
    public function get(string $url, array $parameters = [])
    {
        return $this->call('GET', $url, $parameters);
    }

    /**
     * Perform a GET http request.
     *
     * @param  mixed[]  $parameters
     * @return mixed
     * @throws ApiException
     */
    public function post(string $url, array $parameters = [])
    {
        return $this->call('POST', $url, $parameters);
    }

    /**
     * @param string $url
     * @param mixed[] $parameters
     * @return mixed
     * @throws ApiException
     */
    private function call(string $method, string $url, array $parameters)
    {
        $fp = @ fopen('php://temp', 'w+') ?: null;

        $paramOption = $method === 'GET' ? 'query' : 'form_params';

        try {
            $response = $this->client($fp)
                ->request($method, $url, [
                    $paramOption => $parameters,
                    'debug'      => $fp
                ]);

            return json_decode((string) $response->getBody());
        } catch (BadResponseException $e) {
            $response = json_decode((string) $e->getResponse()->getBody());
            if (isset($response->status) && $response->status === 'error' && ! empty($response->message)) {
                throw new ApiException($response->message, $e->getCode(), $e);
            }

            throw new ApiException($e->getMessage(), $e->getCode(), $e);
        } catch (TransferException $e) {
            throw new ApiException($e->getMessage(), $e->getCode(), $e);
        } finally {
            if ($fp) {
                @ fclose($fp);
            }
        }
    }

    protected function connectionTimeout(): int
    {
        return (int) Config::get('app.marketplace.connect_timeout', 10);
    }

    protected function timeout(): int
    {
        return (int) Config::get('app.marketplace.timeout', 10);
    }

    /**
     * @param resource|null $stream
     */
    private function client($stream): Client
    {
        $stack = HandlerStack::create();
        $stack->push(Middleware::log($this->logger, new MessageFormatter($stream), 'debug'));
        $stack->push(new CacheMiddleware(
            new PublicCacheStrategy(
                new LaravelCacheStorage(Cache::store())
            )
        ), 'cache');

        return $this->proxyService->createGuzzle([
            'handler'         => $stack,
            'base_uri'        => $this->getBaseUrl(),
            'connect_timeout' => $this->connectionTimeout(),
            'timeout'         => $this->timeout(),
            'headers' => [
                'User-Agent'        => 'SupportPal Marketplace Client (https://www.supportpal.com)',
                'Accept'            => 'application/json',
                'X-Product-Version' => Config::get('release.version'),
            ]
        ]);
    }
}
