<?php declare(strict_types=1);

namespace SupportPal\Support\Database;

use Illuminate\Database\MySqlConnection;
use Illuminate\Support\Facades\Cache;
use stdClass;

use function array_column;
use function now;
use function rtrim;
use function sprintf;

/**
 * @mixin MySqlConnection
 */
class MySqlConnectionMacros
{
    /**
     * Get the max allowed packet size for the MySQL Server in Kilobytes.
     */
    public function maxPacketSize(): callable
    {
        return function (): int {
            return (int) Cache::remember('max_allowed_packet', now()->addMinutes(86400), function () {
                $query = $this->select('SELECT @@max_allowed_packet as max_allowed_packet');

                $max = array_column($query, 'max_allowed_packet')[0] ?? 1048576;

                return $max / 1024; // Convert bytes to KB
            });
        };
    }

    /**
     * Get all table columns that are not using the utf8mb4_unicode_ci collation.
     */
    public function getNonUtf8mb4CollationColumns(): callable
    {
        return function (): array {
            $query = 'SELECT * FROM information_schema.COLUMNS WHERE table_schema = ? AND collation_name != ?';

            return $this->select($query, [$this->getDatabaseName(), 'utf8mb4_unicode_ci']);
        };
    }

    /**
     * Get all tables that are not using the utf8mb4_unicode_ci collation.
     */
    public function getNonUtf8mb4CollationTables(): callable
    {
        return function (): array {
            $query = 'SELECT * FROM information_schema.TABLES WHERE table_schema = ? AND table_collation != ?';

            return $this->select($query, [$this->getDatabaseName(), 'utf8mb4_unicode_ci']);
        };
    }

    /**
     * Get all table columns that are not using the utf8mb4 character set.
     */
    public function getNonUtf8mb4CharsetColumns(): callable
    {
        return function (): array {
            $query = 'SELECT * FROM information_schema.COLUMNS WHERE table_schema = ? AND character_set_name != ?';

            return $this->select($query, [$this->getDatabaseName(), 'utf8mb4']);
        };
    }

    /**
     * Convert information_schema output to a comma delimited list (use in combination with
     * getNonUtf8mb4CharsetColumns, getNonUtf8mb4CollationTables, getNonUtf8mb4CollationColumns).
     */
    public function schemaToString(): callable
    {
        return function (array $records): string {
            $string = '';
            /** @var stdClass $record */
            foreach ($records as $record) {
                if (isset($record->COLUMN_NAME)) {
                    $string .= sprintf('%s.%s, ', $record->TABLE_NAME, $record->COLUMN_NAME);
                } else {
                    $string .= sprintf('%s, ', $record->TABLE_NAME);
                }
            }

            return rtrim($string, ', ');
        };
    }

    /**
     * Get list of InnoDB full text stop words.
     */
    public function getStopWords(): callable
    {
        return function (): array {
            return (array) Cache::rememberForever('fulltext_stopwords', function () {
                $query = $this->select('SELECT * FROM INFORMATION_SCHEMA.INNODB_FT_DEFAULT_STOPWORD');

                return array_column($query, 'value');
            });
        };
    }

    /**
     * Get the InnoDB full text minimum token size.
     */
    public function getFtsMinTokenSize(): callable
    {
        return function (): int {
            return (int) Cache::rememberForever('fulltext_min_token_size', function () {
                $size = $this->select('SELECT @@innodb_ft_min_token_size AS size');

                return array_column($size, 'size')[0] ?? 0;
            });
        };
    }
}
